/*
 * Author: Daniel Pavel <daniel.pavel@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name of the Softwin nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package sk.lacike.ntlm;

import java.io.IOException;
import java.util.Iterator;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


import com.atlassian.seraph.auth.Authenticator;
import com.atlassian.seraph.auth.AuthenticatorException;
import com.atlassian.seraph.auth.DefaultAuthenticator;
import com.atlassian.seraph.interceptor.LoginInterceptor;
import com.atlassian.seraph.filter.LoginFilter;


/**
 * The NTLM filter.
 * This class is derived from the standard JIRA login filter because we sometimes
 * need to fallback on the standard (form) way of authenticating.
 */
public class JIRALoginFilter extends LoginFilter {
	private static final Log LOG = LogFactory.getLog(JIRALoginFilter.class);

	// used by JIRA
	public static final String OS_DESTINATION = "os_destination";
	public static final String OS_USERNAME = "os_username";
	public static final String OS_COOKIE = "os_cookie";


	private static Authenticator replacementAuthenticator = null;

	/**
	 * The dashboard URL, read from the JIRA security configuration.
	 */
	private static String dashboard = null;

	/*
	public void init(final FilterConfig config) {
		super.init(config);
	}
	*/

	public String getDashboardURL() {
		if (dashboard == null) {
			dashboard = getSecurityConfig().getLinkLoginURL();

			final int questionMark = dashboard.indexOf('?');
			if (questionMark > -1) {
				dashboard = dashboard.substring(1, questionMark);
			}

			if (dashboard.length() > 0 && dashboard.charAt(0) != '/') {
				dashboard = "/" + dashboard;
			}

			LOG.debug("Using dashboard: " + dashboard);
		}

		return dashboard;
	}


	/**
	 * Check if we need to do any redirects, and apply them.
	 * This method could probably migrate to super.
	 * @return true if any redirect has been applied
	 */
	protected boolean checkRedirects(final HttpServletRequest req, final HttpServletResponse res) throws IOException {
		final String context = req.getContextPath();

		final String originalURL = (String) req.getSession().getAttribute(getSecurityConfig().getOriginalURLKey());
		if (originalURL != null) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("Logged In - Redirecting to Original URL: " + context + originalURL);
			}

			req.getSession().setAttribute(getSecurityConfig().getOriginalURLKey(), null);
			res.sendRedirect(context + originalURL);
			return true;
		}

		final String destination = req.getParameter(OS_DESTINATION);
		if (destination != null) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("Logged In - redirecting to os_destination: " + context + destination);
			}

			res.sendRedirect(context + destination);
			return true;
		}

		return false;
	}


	/**
	 * Perform the actual JIRA login.
	 * This particular piece of code could probably migrate to super...
	 */
	protected boolean doLogin(final HttpServletRequest _req, final HttpServletResponse _res, final String _username, final String _password) {
		if (LOG.isDebugEnabled()) {
			LOG.debug("starting JIRA login");
		}

		// finally
		final List interceptors = getSecurityConfig().getInterceptors(LoginInterceptor.class);
		LoginInterceptor loginInterceptor;
		String result = LOGIN_ERROR;

		for (final Iterator iterator = interceptors.iterator(); iterator.hasNext();
				loginInterceptor.beforeLogin(_req, _res, _username, _password, false)) {
			loginInterceptor = (LoginInterceptor) iterator.next();
		}

		boolean loggedIn = false;
		try {
			final Authenticator auth = getAuthenticator();
			loggedIn = auth.login(_req, _res, _username, _password, false);
			result = loggedIn ? LOGIN_SUCCESS : LOGIN_FAILED;
		} catch (AuthenticatorException aex) {
			LOG.error("jira logon error", aex);
		}

		LOG.debug("Login: " + result + " (" + _username + ")");

		_req.setAttribute(OS_AUTHSTATUS_KEY, result);

		postAuthentication(loggedIn);

		for (final Iterator iterator = interceptors.iterator(); iterator.hasNext();
				loginInterceptor.afterLogin(_req, _res, _username, _password, false, result)) {
			loginInterceptor = (LoginInterceptor) iterator.next();
		}

		return loggedIn;
	}


	protected void postAuthentication(final boolean _success) {
		// nothing to do?
	}


	protected Authenticator getAuthenticator() {
		if (replacementAuthenticator == null) {
			replacementAuthenticator = super.getAuthenticator();

			if (replacementAuthenticator.getClass() == DefaultAuthenticator.class) {
				replacementAuthenticator = new BypassDefaultAuthenticator(getSecurityConfig());
			} else {
				if (LOG.isErrorEnabled()) {
					LOG.error("Non-standard JIRA/seraph configuration, cannot use DefaultAuthenticator bypasser.");
				}
			}
		}

		return replacementAuthenticator;
	}
}
