/*
 * Author: Daniel Pavel <daniel.pavel@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name of the Softwin nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package sk.lacike.ntlm;

import javax.naming.Context;
import java.util.Properties;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


import com.atlassian.core.user.GroupUtils;
import com.atlassian.core.user.UserUtils;
import com.atlassian.core.util.RandomGenerator;
import com.opensymphony.user.DuplicateEntityException;
import com.opensymphony.user.EntityNotFoundException;
import com.opensymphony.user.Group;
import com.opensymphony.user.ImmutableException;
import com.opensymphony.user.User;
import com.opensymphony.user.UserManager;


public final class JIRAUserUtils {
	private static final Log LOG = LogFactory.getLog(JIRAUserUtils.class);

	private JIRAUserUtils() {
		// disabled
	}

	// We autogenerate passwords this long
	private static final int PASSWORD_LENGTH = 16;

	/**
	 *
	 */
	public static User getUser(final String _domain, final String _username) {
		if (LDAPConfig.isCheckLDAP()) {
			final Properties conf = LDAPConfig.getConfig(_domain);
			if (conf == null) {
				throw new IllegalStateException("no configuration found for " + _domain);
			}
	
			final String ldapRequired = conf.getProperty(LDAPConfig.REQUIRED_LDAP_GROUP);
			LDAPHelper helper = null;
			if (ldapRequired != null) {
				 helper = LDAPHelper.create(conf);
				 if (helper == null) {
					 return null;
				 }
			}
	
			final boolean autocreate = LDAPConfig.isTrue(conf, LDAPConfig.USER_AUTOCREATE);
			final String jiraRequiredGroup = conf.getProperty(LDAPConfig.REQUIRED_JIRA_GROUP);
			final Group jiraRequired = jiraRequiredGroup == null ? null : GroupUtils.getGroup(jiraRequiredGroup);
	
			final User user = getUser(_domain, _username, autocreate, jiraRequired, helper, ldapRequired);
			if (helper != null) {
				helper.close();
			}
	
			return user;
		} else {
			final User user = getJiraUser(_domain, _username);
	
			return user;			
		}
	}

	/**
	 *
	 */
	public static User getUser(final String _domain, final String _username, final boolean _autocreate, final Group _jiraRequired) {
		return JIRAUserUtils.getUser(_domain, _username, _autocreate, _jiraRequired, null, null);
	}
	
	public static User getJiraUser(final String _domain, final String _username) {

		User user = null;
		try {
			user = UserManager.getInstance().getUser(_username);
		} catch (EntityNotFoundException enfex) {
			LOG.debug("User " + _username + " from domain " + _domain + " not found");
			return null;
		}
		
		return user;
		
	}

	public static User getUser(final String _domain, final String _username, final boolean _autocreate, final Group _jiraRequired, final LDAPHelper _helper, final String _ldapRequiredGroup) {
		final String username = _username.toLowerCase();

		// get the user, check JIRA permission
		User user = getJiraUser(_domain, _username);
		// user not found and autocreate is disabled
		if (user == null && !_autocreate)
			return null;
		// check required group
		if (user != null && _jiraRequired != null && !user.inGroup(_jiraRequired)) {
			LOG.warn("denied access to " + username + ", not member of JIRA group " + _jiraRequired);
			return null;
		}
		
		// check LDAP permission
		if (_helper != null && !_helper.isMemberOfGroup(username, _ldapRequiredGroup)) {
			LOG.warn("denied access to " + username + ", not member of LDAP group " + _ldapRequiredGroup);
			return null;
		}

		if (user == null) {
			// create the user
			LOG.warn("will create JIRA user " + username + " from (" + _domain + "\\" + username + ")");
			try {
				user = UserUtils.createUser(username, RandomGenerator.randomString(PASSWORD_LENGTH), "", null);

				if (_jiraRequired != null) {
					user.addToGroup(_jiraRequired);
				}
			} catch (DuplicateEntityException deex) {
				// wtf?
				LOG.error("autocreate user:", deex);
			} catch (ImmutableException iex) {
				// huh?
				LOG.error("autocreate user:", iex);
			}
		}

		return user;
	}


	/**
	 * Gets the user's fullname and email from the configured LDAP.
	 * @return true if the operation went ok (which includes no actual update taking place, _if_ we have no LDAP configuration)
	 */
	public static boolean updateUserData(final User _user, final String _domain, final String _username) {
		
		// readonly user -> report success
		if (LDAPConfig.isReadonlyUser())
			return true;

		// this requires ldap access
		final Properties conf = LDAPConfig.getConfig(_domain);
		if (conf == null) {
			throw new IllegalStateException("no configuration found for " + _domain);
		}

		final String url = conf.getProperty(Context.PROVIDER_URL);
		if (url == null) { // no LDAP configuration
			return true;
		}

		LOG.debug("updating " + _user + " from " + url);
		final LDAPHelper helper = LDAPHelper.create(conf);
		if (helper == null) {
			return false;
		}

		final String dn = helper.getUserDN(_username);
		if (dn == null) {
			helper.close();
			throw new IllegalStateException("no user " + _username + " found in LDAP for " + _domain);
		}

		final String[] attributeNames = new String[] {
				conf.getProperty(LDAPConfig.EMAIL_ATTRIBUTE),
				conf.getProperty(LDAPConfig.FULLNAME_ATTRIBUTE)
			};
		final String[] nameAndEmail = helper.getAttributes(dn, attributeNames);
		
		
		if (nameAndEmail.length == 2) {
			if (!LDAPConfig.isReadonlyUser()) {
				if (nameAndEmail[0] != null) {
					_user.setEmail(nameAndEmail[0]);
				}
				if (nameAndEmail[1] != null) {
					_user.setFullName(nameAndEmail[1]);
				}
			} else {
				LOG.debug("updating " + _user + " from " + url + " is not allowed - user provider is read only");
			}
		}

		helper.close();
		return true;
	}


	/**
	 * Generate a new random password for the given user.
	 * @return null if we can't modify the user, the new password otherwise
	 */
	public static String newPassword(final User _user) {
		final String password = RandomGenerator.randomString(PASSWORD_LENGTH);
		try {
			_user.setPassword(password);
		} catch (ImmutableException iex) {
			LOG.error("newPassword for " + _user, iex);
			return null;
		}
		return password;
	}
}
