/*
 * Author: Daniel Pavel <daniel.pavel@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name of the Softwin nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package sk.lacike.ntlm;

import java.io.IOException;
import java.io.InputStream;
import java.net.UnknownHostException;

import javax.naming.Context;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import sk.lacike.ntlm.jcifs.jira.NTLMWebHelper;
import sk.lacike.ntlm.jcifs.jira.SMBHelper;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;


import com.opensymphony.util.FileManager;


public class LDAPConfig {
	private static final Log LOG = LogFactory.getLog(LDAPConfig.class);

	//
	public static final String DOMAINS = "domains";
	public static final String DOMAIN_CONTROLLER = "domainController";
	/* lacike@lacike.sk BEGIN */
	public static final String STRIP_DOMAIN = "stripDomain";
	public static final String FORCE_DOMAIN = "forceDomain";
	protected static String READONLY_USER = "readonlyUser";
	protected static String CHECK_LDAP = "checkLDAP";
	/* lacike@lacike.sk END */

	public static final String USER_AUTOCREATE = "autoCreate";
	public static final String REQUIRED_JIRA_GROUP = "requiredJIRAGroup";
	public static final String REQUIRED_LDAP_GROUP = "requiredLDAPGroup";

	public static final String EMAIL_ATTRIBUTE = "emailAttribute";
	public static final String FULLNAME_ATTRIBUTE = "fullNameAttribute";
	public static final String GROUP_MEMBER_ATTRIBUTE = "groupMemberAttribute";

	// keeps a Properties object for each domain
	protected static final Map CONFIGURATIONS = new HashMap();
	/* lacike@lacike.sk holds value if domain should be stripped BEGIN */
	protected static boolean STRIP_DOMAIN_VALUE = false;
	protected static String FORCE_DOMAIN_VALUE = null;
	protected static boolean READONLY_USER_VALUE = false;
	protected static boolean CHECK_LDAP_VALUE = true;
	/* lacike@lacike.sk END */

	/**
	 * Gets the configuration map for the given domain, if available.
	 */
	public static Properties getConfig(final String _domain) {
		if (_domain == null) {
			return null;
		}

		return (Properties) CONFIGURATIONS.get(_domain);
	}


	/**
	 * @return true if the key has been set to any value in the destination map
	 */
	public static boolean setProperty(final Properties _dst, final String _key, final Properties _src, final String _default) {
		if (_dst == null || _src == null || _key == null) {
			return false;
		}

		String value = null;
		final String domain = _dst.getProperty(DOMAINS);
		if (domain != null) {
			value = _src.getProperty(domain + "." + _key);
			if (value == null) {
				value = _src.getProperty(domain.toLowerCase() + "." + _key);
			}
			if (value == null) {
				value = _src.getProperty(domain + "." + _key.toLowerCase());
			}
			if (value == null) {
				value = _src.getProperty(domain.toLowerCase() + "." + _key.toLowerCase());
			}
		}

		if (value == null) {
			value = _src.getProperty(_key);
		}
		if (value == null) {
			value = _src.getProperty(_key.toLowerCase());
		}

		if (value == null) {
			value = _default;
		}

		if (value != null) {
			value = value.trim();
			// we're being nice with quotes
			if (value.length() > 1
					&& ((value.charAt(0) == '"' && value.endsWith("\""))
							|| (value.charAt(0) == '\'' && value.endsWith("'")))) {
				value = value.substring(1, value.length() - 1);
			}
			if (value.length() < 1) {
				value = null;
			}
		}

		if (value == null) {
			_dst.remove(_key);
			return false;
		}

		_dst.setProperty(_key, value);
		return true;
	}


	/**
	 * @return true if the given property is set to either 'true' or 'yes'
	 */
	public static boolean isTrue(final Properties _properties, final String _key) {
		final String value = _properties.getProperty(_key);
		if (value == null) {
			return false;
		}

		return "true".equalsIgnoreCase(value) || "yes".equalsIgnoreCase(value) || "on".equalsIgnoreCase(value);
	}


	/**
	 * @return true if a valid configuration has been read from the given map
	 */
	public static boolean readConfiguration(final Properties _properties) {
		CONFIGURATIONS.clear();

		final String domainsProperty = _properties.getProperty(DOMAINS);
		if (domainsProperty == null) {
			LOG.warn("no 'domains' property found in ntlm configuration");
			return false;
		}

		final String controller = _properties.getProperty(DOMAIN_CONTROLLER);
		try {
			SMBHelper.setDC(controller);
		} catch (UnknownHostException uhex) {
			LOG.error("cannot use DC '" + controller + "'", uhex);
		}

		if (SMBHelper.getDC() == null) {
			LOG.error("no Domain Controller available!");
			return false;
		}
		
		/* lacike@lacike.sk Parse property determining if domain should be stripped BEGIN */
		final String stripDomainAsStr = _properties.getProperty(STRIP_DOMAIN);
		if (stripDomainAsStr != null) {
			try {
				STRIP_DOMAIN_VALUE = Boolean.parseBoolean(stripDomainAsStr);
			} catch (Exception e) {
				LOG.error("unable to parse boolean value of " + STRIP_DOMAIN + " = '" + stripDomainAsStr + "'", e);				
			}
		} else
			STRIP_DOMAIN_VALUE = false;

		// read only user information
		final String readonlyUserAsStr = _properties.getProperty(READONLY_USER);
		if (readonlyUserAsStr != null) {
			try {
				READONLY_USER_VALUE = Boolean.parseBoolean(readonlyUserAsStr);
			} catch (Exception e) {
				LOG.error("unable to parse boolean value of " + READONLY_USER + " = '" + readonlyUserAsStr + "'", e);				
			}
		} else
			READONLY_USER_VALUE = false;
		
		// Forced domain
		FORCE_DOMAIN_VALUE = _properties.getProperty(FORCE_DOMAIN);
		if (FORCE_DOMAIN_VALUE == null || FORCE_DOMAIN_VALUE.trim().isEmpty())
			FORCE_DOMAIN_VALUE = null;

		// read only user information
		final String checkLDAPUserAsStr = _properties.getProperty(CHECK_LDAP);
		if (checkLDAPUserAsStr != null) {
			try {
				CHECK_LDAP_VALUE = Boolean.parseBoolean(checkLDAPUserAsStr);
			} catch (Exception e) {
				LOG.error("unable to parse boolean value of " + CHECK_LDAP + " = '" + checkLDAPUserAsStr + "'", e);				
			}
		} else
			CHECK_LDAP_VALUE = true;
		
		/* lacike@lacike.sk END */

		final String[] domains = domainsProperty.toUpperCase().split(",");
		for (int i = 0; i < domains.length; i++) {
			String domain = domains[i].trim();
			if (domain.length() < 1) {
				continue;
			}

			final Properties domainConfig = new Properties();

			if (domain.indexOf('.') > -1) {
				domain = domain.substring(0, domain.indexOf('.'));
			}
			domainConfig.setProperty(DOMAINS, domain);

			setProperty(domainConfig, USER_AUTOCREATE, _properties, null);
			setProperty(domainConfig, REQUIRED_JIRA_GROUP, _properties, null);

			if (setProperty(domainConfig, Context.PROVIDER_URL, _properties, null)) {
				if (!setProperty(domainConfig, LDAPHelper.SEARCH_BASE, _properties, null)) {
					domainConfig.remove(Context.PROVIDER_URL);
					LOG.error("configuration missing ldap .searchBase for domain " + domain);
					continue;
				}

				setProperty(domainConfig, Context.SECURITY_AUTHENTICATION, _properties, "simple");
				setProperty(domainConfig, Context.INITIAL_CONTEXT_FACTORY, _properties, "com.sun.jndi.ldap.LdapCtxFactory");
				setProperty(domainConfig, Context.REFERRAL, _properties, "follow");

				if (setProperty(domainConfig, Context.SECURITY_PRINCIPAL, _properties, null)) {
					setProperty(domainConfig, Context.SECURITY_CREDENTIALS, _properties, null);
				} else {
					LOG.warn("searches in " + domainConfig.getProperty(Context.PROVIDER_URL) + " for domain " + domain + " will be done anonymously");
				}

				// the defaults are sensible for a Windows Active Directory
				setProperty(domainConfig, LDAPHelper.UID_ATTRIBUTE, _properties, "sAMAccountName");
				setProperty(domainConfig, FULLNAME_ATTRIBUTE, _properties, "displayName");
				setProperty(domainConfig, EMAIL_ATTRIBUTE, _properties, "mail");
				setProperty(domainConfig, GROUP_MEMBER_ATTRIBUTE, _properties, "memberOf");

				setProperty(domainConfig, REQUIRED_LDAP_GROUP, _properties, null);
			} else {
				LOG.warn("no '" + Context.PROVIDER_URL + "' for domain " + domain + "; user data will not be available");
			}

			if (CONFIGURATIONS.get(domain) != null) {
				LOG.warn("overrding previous configuration for domain " + domain);
			}
			CONFIGURATIONS.put(domain, domainConfig);

			LOG.info("domain " + domain + " using " + domainConfig.getProperty(Context.PROVIDER_URL) + " : " + domainConfig.getProperty(LDAPHelper.SEARCH_BASE));
		}

		if (CONFIGURATIONS.size() < 1) {
			LOG.error("No properly configured domains found!");
			return false;
		}

		if (CONFIGURATIONS.size() == 1) {
			// when only one domain configured, we allow the users to enter just the username
			NTLMWebHelper.defaultDomain = CONFIGURATIONS.keySet().iterator().next().toString();
		}

		return true;
	}


	/**
	 * @return true if a valid configuration has been read from the given file
	 */
	public static boolean read(final String _filename) {
		if (_filename == null) {
			LOG.error("No file to read configuration from");
			return false;
		}

		InputStream input = null;
		try {
			// this gives us JIRA restart when the file changes
			input = FileManager.loadFile(_filename, LDAPConfig.class);
		} catch (Exception ex) {
			LOG.error("Could not load configuration from file " + _filename, ex);
			return false;
		}

		if (input == null) {
			LOG.error("Missing configuration file " + _filename);
			return false;
		}

		LOG.info("Loading configuration from " + _filename);
		final Properties ldapConfig  = new Properties();
		try {
			ldapConfig.load(input);
		} catch (Exception ex) {
			LOG.error("failed loading configuration", ex);
			return false;
		} finally {
			try { input.close(); } catch (IOException ioex) {
				// best effort...
			}
		}

		return readConfiguration(ldapConfig);
	}


	public static boolean allowDomain(final String _domain) {
		final Properties conf = getConfig(_domain);
		return conf != null;
	}
	
	public static boolean isStripDomain() {
		return STRIP_DOMAIN_VALUE;
	}
	
	public static boolean isForceDomain() {
		return FORCE_DOMAIN_VALUE != null;
	}
	
	public static String getForceDomain() {
		return FORCE_DOMAIN_VALUE;
	}
	
	public static boolean isReadonlyUser() {
		return READONLY_USER_VALUE;
	}
	
	public static boolean isCheckLDAP() {
		return CHECK_LDAP_VALUE;
	}
}
