/*
 * Author: Daniel Pavel <daniel.pavel@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name of the Softwin nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */


package sk.lacike.ntlm;

import java.util.Properties;

import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchControls;
import javax.naming.directory.SearchResult;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;



/**
 * Some LDAP-related functions to make common operations less painful.
 *
 * In case the naming API throws anything, we log the exceptions in the JIRA log,
 * and return some value suggesting error.  This is useful because in the servlet filter
 * we just want to know if it worked fine or not; in case of exception there is no point
 * in dumping it to the user, we just let him know it didn't work and the site admin
 * can check what's wrong (because, mostly, he's the only one who can fix it anyway).
 */
public class LDAPHelper {
	private static final Log LOG = LogFactory.getLog(LDAPHelper.class);

	/**
	 * Attempts to create a LDAP connection as described in the given properties map.
	 * @return a LDAP helper instance if it worked, null otherwise
	 */
	public static LDAPHelper create(final Properties _env) {
		try {
			return new LDAPHelper(_env);
		} catch (NamingException nex) {
			LOG.error("Could not connect to LDAP", nex);
			return null;
		}
	}

	public static final String UID_ATTRIBUTE = "uidAttribute";
	public static final String SEARCH_BASE = "searchBase";
	public static final String[] EMPTY = new String[] {};


	///////////////////////////////////////

	private final InitialDirContext context;
	private final String[] uidAttribute;
	private final String searchBase;
	private final String groupMember;


	/**
	 * Tries to init a Naming Context with the given properties.
	 * The properties map needs to have two additional keys set (besides the standard javax.naming ones),
	 * "uidAttribute" (the name of the attribute that specifies a user's id) and "searchBase"
	 * (the LDAP base DN for searching).
	 * A third property, "groupMemberAttribute" is used by the isMemberOfGroup method.
	 * @throws NamingException
	 */
	public LDAPHelper(final Properties _env) throws NamingException {
		context = new InitialDirContext(_env);

		uidAttribute = new String[] { _env.getProperty(UID_ATTRIBUTE) };
		searchBase = _env.getProperty(SEARCH_BASE);
		groupMember = _env.getProperty(LDAPConfig.GROUP_MEMBER_ATTRIBUTE);
		LOG.debug("Inited LDAP context " + searchBase);
	}


	/**
	 * Closes this connection.  Should help with the cleanup, so it's strongly suggested.
	 */
	public void close() {
		LOG.debug("closing LDAP context " + searchBase);
		try {
			context.close();
		} catch (NamingException nex) {
			LOG.error("closing LDAP connection", nex);
		}
	}


	/**
	 * Finds the DN for the given LDAP user, by searching for the uid attribute.
	 * @return null if some error occurs
	 */
	public String getUserDN(final String _username) {
		LOG.debug("searching for user '" + _username + "' in " + searchBase);
		final SearchControls sc = new SearchControls(SearchControls.SUBTREE_SCOPE, 1, 0, uidAttribute, false, false);
		
		NamingEnumeration results = null;
		try {
			results = context.search(searchBase, uidAttribute[0] + "=" + _username, sc);

			if (results == null || !results.hasMore()) {
				LOG.warn("'" + _username + "' not found in " + searchBase);
				return null;
			}

			final SearchResult sr = (SearchResult) results.next();
			return sr.getName() + "," + searchBase;
		} catch (NamingException nex) {
			LOG.error("Could not LDAP search for " + uidAttribute[0] + "='" + _username + "'", nex);
			return null;
		} finally {
			if (results != null) {
				try {
					results.close();
				} catch (NamingException nex) {
					// best effort
				}
			}
		}
	}


	/**
	 * Gets values for a number of attributes from a specific DN.
	 * In case of multi-value attributes, only one value is returned.
	 * @return an array of values for the attributes, or an empty array in case of error
	 */
	public String[] getAttributes(final String _dn, final String[] _attributeNames) {
		LOG.debug("reading attributes for " + _dn + " " + _attributeNames);
		Attributes attrs = null;

		try {
			attrs = context.getAttributes(_dn, _attributeNames);
		} catch (NamingException nex) {
			LOG.error("Could not read attributes for '" + _dn + "'", nex);
			return EMPTY;
		}

		final int count = _attributeNames.length;
		final String[] values = new String[count];

		try {
			for (int i = 0; i < count; i++) {
				final Attribute a = attrs.get(_attributeNames[i]);
				final Object v = a == null ? null : a.get();
				values[i] = v == null ? null : v.toString();
			}
		} catch (NamingException nex) {
			LOG.error("Error packing attribute values", nex);
			return EMPTY;
		}

		return values;
	}


	/**
	 * Fetches the given attribute for a DN.
	 * @return an array of values for that attribute, or an empty array if some error occured
	 */
	public String[] getAttribute(final String _dn, final String _attribute) {
		LOG.debug("reading '"  + _attribute + "' for " + _dn);
		Attributes attrs = null;
		try {
			attrs = context.getAttributes(_dn, new String[] { _attribute });
		} catch (NamingException nex) {
			LOG.error("Could not read attributes for '" + _dn + "'", nex);
			return EMPTY;
		}

		if (attrs == null) {
			return EMPTY;
		}
		final Attribute groups = attrs.get(_attribute);
		if (groups == null) {
			return EMPTY;
		}
		final String[] values = new String[groups.size()];

		try {
			for (int i = 0; i < values.length; i++) {
				Object v = groups.get(i);
				values[i] = v == null ? "" : v.toString();
			}
		} catch (NamingException nex) {
			LOG.error("Error packing attribute values", nex);
			return EMPTY;
		}

		return values;
	}


	/**
	 * Checks if a LDAP user is a member of the given LDAP group.
	 * The LDAP group can be either specified as full DN, or just the CN.
	 * @return true if the user exists and is a member of that group
	 */
	public boolean isMemberOfGroup(final String _username, final String _ldapGroup) {
		String userDN = getUserDN(_username);
		if (userDN == null) {
			return false;
		}

		final String[] groups = getAttribute(userDN, groupMember);

		final boolean fullDN = _ldapGroup.startsWith("CN=");
		final String ldapGroup = fullDN ? _ldapGroup : "CN=" + _ldapGroup + ",";

		boolean found = false;
		for (int i = 0; i < groups.length && !found; i++) {
			if (fullDN) {
				found = groups[i].equals(ldapGroup);
			} else {
				found = groups[i].startsWith(ldapGroup);
			}
		}

		if (!found) {
			LOG.warn(_username + " is not a member of " + _ldapGroup);
		}

		return found;
	}
}
