package sk.lacike.ntlm.apache.jira;

import java.io.IOException;
import java.security.Principal;

import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.apache.log4j.Category;

import com.opensymphony.user.User;
import com.atlassian.seraph.config.SecurityConfig;

import sk.lacike.ntlm.BypassDefaultAuthenticator;
import sk.lacike.ntlm.JIRALoginFilter;
import sk.lacike.ntlm.JIRAUserUtils;
import sk.lacike.ntlm.LDAPConfig;

//
// here's the beef
//

/**
 * The NTLM filter.
 * This class is derived from the standard JIRA login filter because we sometimes
 * need to fallback on the standard (form) way of authenticating.
 */
public class AJPLoginFilter extends JIRALoginFilter {
	private static final Category LOG = Category.getInstance(AJPLoginFilter.class);

	public static final String AJP_AUTH_VERSION = "0.1";

	// This filter has one parameter named "configuration", whose value
	// points to the file in WEB-INF/classes that contains the actual configuration :).
	public static final String LDAP_CONFIG_PARAM = "configuration";

	/**
	 * The whole NTLM thingy can be disabled just like that.
	 */
	public static boolean isNTLMEnabled = false;

	public void init(final FilterConfig config) {
		super.init(config);

		isNTLMEnabled = LDAPConfig.read(config.getInitParameter(LDAP_CONFIG_PARAM));
		if (!isNTLMEnabled) {
			LOG.error("NTLM authentication disabled.");
		}
	}

	protected static String getRemoteUser(final HttpServletRequest _req, final HttpServletResponse _res) {
		final SecurityConfig securityConfig = (SecurityConfig) _req.getSession().getServletContext().getAttribute("seraph_config");
		final Principal user = securityConfig.getAuthenticator().getUser(_req, _res);

		if (user != null && user.getName() != null && !user.getName().isEmpty())
			return user.getName();
		
		return _req.getRemoteUser();
	}
	
	public void doFilter(final ServletRequest req, final ServletResponse res, final FilterChain chain)
		throws IOException, ServletException {

		final HttpServletRequest hreq = (HttpServletRequest) req;
		final HttpServletResponse hres = (HttpServletResponse) res;

		// is this already filtered?
		boolean alreadyFiltered = hreq.getAttribute(ALREADY_FILTERED) != null;
		if (alreadyFiltered || !getSecurityConfig().getController().isSecurityEnabled()) {
			LOG.debug("request already filtered (" + hreq.getServletPath() + ")");
			
			chain.doFilter(req, res);			
			return;
		}

		// get remote user from Apache HTTP server
		String remoteUser = getRemoteUser(hreq, hres);
		LOG.debug("Provided username is " + remoteUser);
		
		hreq.setAttribute(ALREADY_FILTERED, Boolean.TRUE);
		hreq.setAttribute(OS_AUTHSTATUS_KEY, LOGIN_NOATTEMPT);			

		// What? No username?
		if (remoteUser == null || remoteUser.isEmpty()) {
			LOG.error("No username provided, authentication failed.");
			HttpSession ssn = hreq.getSession(true);
			ssn.setAttribute(OS_AUTHSTATUS_KEY, LOGIN_FAILED);
			req.setAttribute(OS_AUTHSTATUS_KEY, LOGIN_FAILED);
		} else {
			// try to log in
			boolean loggedIn = doLogin(hreq, hres, "", remoteUser);
	
			HttpSession ssn = hreq.getSession(true);
			
			// Could not find user, probably permission issue..
			if (!loggedIn) {
				LOG.error(remoteUser + ": NTLM went ok but could not log in.");
				ssn.setAttribute(OS_AUTHSTATUS_KEY, LOGIN_FAILED);
				req.setAttribute(OS_AUTHSTATUS_KEY, LOGIN_FAILED);
			// successfully filtered
			} else {
				ssn.setAttribute(OS_AUTHSTATUS_KEY, LOGIN_SUCCESS);			
				req.setAttribute(OS_AUTHSTATUS_KEY, LOGIN_SUCCESS);
			}
	
			if (loggedIn && checkRedirects(hreq, hres)) {
				return;
			}
		}
			
		// whatever the outcome, we pass on to next filter
		chain.doFilter(hreq, hres);
	}

	/**
	 * Attempt a login, making sure the user is created first if necessary _and_ requested via config.
	 */
	public boolean doLogin(final HttpServletRequest _req, final HttpServletResponse _res, final String _domain, final String _username) {
		// We make sure the user exists in the local Jira storage.
		User user = JIRAUserUtils.getUser(_domain, _username);
		if (user == null) {
			LOG.error("User " + _username + " not found in JIRA");
			return false;
		}

		// Update user data from the LDAP.
		if (!JIRAUserUtils.updateUserData(user, _domain, _username)) {
			LOG.error("Failed to update user data from LDAP");
			return false;
		}

		return super.doLogin(_req, _res, _username.toLowerCase(), BypassDefaultAuthenticator.NO_PASSWORD);
	}
}
