/*
 * Author: Daniel Pavel <daniel.pavel@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name of the Softwin nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */


package sk.lacike.ntlm.jcifs.jira;

import java.io.IOException;
import java.security.Principal;

import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.opensymphony.user.User;
import com.atlassian.seraph.config.SecurityConfig;

import sk.lacike.ntlm.BypassDefaultAuthenticator;
import sk.lacike.ntlm.JIRALoginFilter;
import sk.lacike.ntlm.JIRAUserUtils;
import sk.lacike.ntlm.LDAPConfig;
import sk.lacike.ntlm.jcifs.jira.NTLMWebHelper.NTLMResult;

//
// here's the beef
//

/**
 * The NTLM filter.
 * This class is derived from the standard JIRA login filter because we sometimes
 * need to fallback on the standard (form) way of authenticating.
 */
public class NTLMLoginFilter extends JIRALoginFilter {
	private static final Log LOG = LogFactory.getLog(NTLMLoginFilter.class);

	public static final String NTLM_AUTH_VERSION = "0.5.3";

	public static final String REFERER = "Referer";
	public static final String VIA = "Via";

	public static final String LOGIN_JSP = "/login.jsp";
	public static final String JIRA_SETUP_ACTION = "/secure/Setup!default.jspa";

	public static final String OS_PASSWORD = "os_password";
	public static final String OS_AUTHTYPE = "os_authType";
	public static final String OS_AUTHTYPE_BASIC = "basic";

	// This filter has one parameter named "configuration", whose value
	// points to the file in WEB-INF/classes that contains the actual configuration :).
	public static final String LDAP_CONFIG_PARAM = "configuration";

	private boolean ntlmDisabledWarningServed = false;

	/**
	 * The whole NTLM thingy can be disabled just like that.
	 */
	public static boolean isNTLMEnabled = false;

	public void init(final FilterConfig config) {
		super.init(config);

		isNTLMEnabled = LDAPConfig.read(config.getInitParameter(LDAP_CONFIG_PARAM));
		if (!isNTLMEnabled) {
			if (LOG.isErrorEnabled()) {
				LOG.error("NTLM authentication disabled.");
			}
		}
	}

	protected static String getRemoteUser(final HttpServletRequest _req, final HttpServletResponse _res) {
		final SecurityConfig securityConfig = (SecurityConfig) _req.getSession().getServletContext().getAttribute("seraph_config");
		final Principal user = securityConfig.getAuthenticator().getUser(_req, _res);

		return user == null ? null : user.getName();
	}

	public void doFilter(final ServletRequest req, final ServletResponse res, final FilterChain chain)
		throws IOException, ServletException {

		final HttpServletRequest hreq = (HttpServletRequest) req;
		final HttpServletResponse hres = (HttpServletResponse) res;

		final String remoteUser = getRemoteUser(hreq, hres);
		if (LOG.isDebugEnabled()) {
			LOG.debug("[" + remoteUser + "] " + hreq.getMethod() + " " + hreq.getRequestURL()
				+ " (" + hreq.getContentLength() + "b, ref=" + hreq.getHeader(REFERER) + ")");
			LOG.trace("UA:" + hreq.getHeader("User-Agent"));
			LOG.trace("requestURL=" + hreq.getRequestURL()
				+ " requestURI=" + hreq.getRequestURI()
				+ " servletPath=" + hreq.getServletPath());
		}

		if (!isNTLMEnabled) {
			if (LOG.isWarnEnabled() && !ntlmDisabledWarningServed) {
				ntlmDisabledWarningServed = true;
				LOG.warn("NTLM disabled -- misconfiguration?  check above for possible hints");
			}

			super.doFilter(req, res, chain);
			return;
		}

		boolean alreadyFiltered = hreq.getAttribute(ALREADY_FILTERED) != null;
		if (alreadyFiltered || !getSecurityConfig().getController().isSecurityEnabled()) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("request already filtered (" + hreq.getServletPath() + ")");
			}
			
			chain.doFilter(req, res);			
			return;
		}

		// first we check that NTLM is needed for the request
		if (!needsNTLM(hreq, hres, remoteUser)) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("NTLM not needed for " + hreq.getServletPath());
			}
			super.doFilter(req, res, chain);
			return;
		}

		// and then if it's actually allowed
		if (!allowsNTLM(hreq)) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("NTLM not allowed for " + hreq.getServletPath());
			}

			chain.doFilter(req, res);
			return;
		}

		if (hreq.getHeader(VIA) != null) {
			// who do you think you're kidding?  NTLM through proxy?
			if (LOG.isWarnEnabled()) {
				LOG.warn("Silly rabbit, trying to do NTLM from [" + hreq.getRemoteHost() + "] through proxy [" + hreq.getHeader("Via") + "]");
			}
			super.doFilter(req, res, chain);
			return;
		}

		hreq.setAttribute(ALREADY_FILTERED, Boolean.TRUE);
		hreq.setAttribute(OS_AUTHSTATUS_KEY, LOGIN_NOATTEMPT);

		final NTLMResult ntlm = doNTLM(req, res, chain);
		if (ntlm == null) {
			return;
		}

		boolean isIErenegociating = NTLMWebHelper.isIERenegociatingPOST(hreq, hres);
		if (isIErenegociating && !ntlm.getUsername().equals(remoteUser)) {
			// whoa! we're switching users here?  is this even remotely acceptable?
			if (LOG.isWarnEnabled()) {
				LOG.warn("IE-renegociation: user switch!! from [" + remoteUser + "] to [" + ntlm.getDomain() + "\\" + ntlm.getUsername() + "]");
			}

			// hightly unlikely, but... who knows?
			// now we need to re-login in Jira as well
			isIErenegociating = false;
		}

		// If it was just IE throwing tantrums, we don't have to do the internal login all over again
		boolean loggedIn = isIErenegociating;

		// WARNING:  domain1\\user1 and domain2\\user1 will be matched to the same internal user!!!
		if (!loggedIn) {
			loggedIn = doLogin(hreq, hres, ntlm.getDomain(), ntlm.getUsername());

			if (!loggedIn) {
				if (LOG.isErrorEnabled()) {
					LOG.error(ntlm.getUsername() + ": NTLM went ok but could log in.  Disabling NTLM for this session.");
				}
				HttpSession ssn = hreq.getSession(true);
				ssn.setAttribute(OS_AUTHSTATUS_KEY, LOGIN_FAILED);
			}
		}

		if (loggedIn && checkRedirects(hreq, hres)) {
			return;
		}

		// whatever the outcome, we pass on to next filter
		chain.doFilter(hreq, hres);
	}

	protected NTLMResult doNTLM(final ServletRequest req, final ServletResponse res, final FilterChain chain)
	throws IOException, ServletException {

		final HttpServletRequest hreq = (HttpServletRequest) req;
		final HttpServletResponse hres = (HttpServletResponse) res;
		final NTLMResult ntlm = NTLMWebHelper.doNTLM(hreq, hres);

		if (ntlm == null) {
			if (LOG.isWarnEnabled()) {
				LOG.warn("failed to initiate NTLM for " + hreq.getServletPath() + ", relegating to default authentication procedure");
			}

			// failed to get the NTLM going, so give up on it and just delegate to the default login procedure
			hreq.removeAttribute(ALREADY_FILTERED);
			hreq.removeAttribute(OS_AUTHSTATUS_KEY);
			super.doFilter(req, res, chain);
			return null;
		}

		// we're barely starting the ntlm process
		if (ntlm.barelyStarted()) {
			return null;
		}

		if (ntlm.hasFailedConnection()) { // could not contact DC
			if (LOG.isInfoEnabled()) {
				LOG.info("SMB connection failed, falling back to standard authentication");
			}

			hreq.setAttribute(OS_AUTHSTATUS_KEY, LOGIN_ERROR);
			chain.doFilter(req, res);
			return null;
		}

		if (!ntlm.hasFinishedHandshake()) { // we're mid-auth (NTLM is a three-step process)
			return null;
		}

		if (ntlm.hasFailedLogin()) { // ntlm successful, login failed
			hreq.setAttribute(OS_AUTHSTATUS_KEY, LOGIN_FAILED);
			if (LOG.isInfoEnabled()) {
				LOG.info("NTLM login failed, falling back to standard authentication");
			}

			// we disable further NTLM attempts this session, lest the user might get their
			// domain account locked from too many bad tries (page, js, css, pictures...)
			final HttpSession ssn = hreq.getSession(true);
			ssn.setAttribute(OS_AUTHSTATUS_KEY, LOGIN_FAILED);

			chain.doFilter(req, res);
			return null;
		}

		if (LOG.isInfoEnabled()) {
			LOG.info("NTLM ok as [" + ntlm.getDomain() + "\\" + ntlm.getUsername() + "] at " + hreq.getServletPath());
		}

		return ntlm;
	}

	/**
	 * Does this request _need_ NTLM?
	 * @param hreq
	 * @return true is NTLM authentication needs to happen on this request
	 */
	protected boolean needsNTLM(final HttpServletRequest hreq, final HttpServletResponse hres, final String _remoteUser) {
		if (NTLMWebHelper.isIERenegociatingPOST(hreq, hres)) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("needsNTLM: IE POST re-negociation");
			}

			return true;
		}

		if (_remoteUser != null) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("already authenticated as [" + _remoteUser + "] at " + hreq.getServletPath());
			}
			return false;
		}

		if (wantsFormLogin(hreq)) {
			// somebody is trying to login the form way
			if (LOG.isDebugEnabled()) {
				LOG.debug("user wants plain form login for " + hreq.getServletPath());
			}
			return false;
		}

		if (!isAuthenticationRequired(hreq)) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("authentication not needed for " + hreq.getServletPath());
			}
			return false;
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("needsNTLM: yes");
		}

		return true;
	}

	/**
	 * Does this request _allow_ NTLM ?
	 * @param hreq
	 * @return true if we allow NTLM negociation on this request
	 */
	protected boolean allowsNTLM(final HttpServletRequest _req) {
		final HttpSession ssn = _req.getSession(false);
		if (ssn != null && LOGIN_FAILED.equals(ssn.getAttribute(OS_AUTHSTATUS_KEY))) {
			_req.setAttribute(ALREADY_FILTERED, Boolean.TRUE);
			_req.setAttribute(OS_AUTHSTATUS_KEY, LOGIN_FAILED);

			if (LOG.isInfoEnabled()) {
				LOG.info(_req.getServletPath() + " session has previously tried NTLM, and failed");
			}
			return false;
		}

		if (LOG.isDebugEnabled()) {
			LOG.debug("allowsNTLM: yes");
		}

		return true;
	}


	protected boolean wantsFormLogin(final HttpServletRequest _req) {
		// Use FormsLogin if password is supplied on the query string
		if (_req.getParameter(OS_PASSWORD) != null) {
			if (LOG.isTraceEnabled()) {
				LOG.trace("request " + _req.getServletPath() + " provided os_password");
			}
			return true;
		}
		if (_req.getParameter(OS_AUTHTYPE) != null && OS_AUTHTYPE_BASIC.equals(_req.getParameter(OS_AUTHTYPE))) {
			if (LOG.isTraceEnabled()) {
				LOG.trace("request " + _req.getServletPath() + " required basic authentication");
			}
			return true;
		}

		final String path = _req.getServletPath();
		if (LOGIN_JSP.equals(path)) {
			if (LOG.isTraceEnabled()) {
				LOG.trace("request " + _req.getServletPath() + " is login.jsp");
			}
			return true;
		}

		return false;
	}


	protected boolean isAuthenticationRequired(final HttpServletRequest _req) {
		final String path = _req.getServletPath();
		if (JIRA_SETUP_ACTION.equals(path)) {
			isNTLMEnabled = false;
			if (LOG.isWarnEnabled()) {
				LOG.warn("Detected JIRA setup procedure, disabling NTLM login");
				LOG.warn("If you want NTLM, you'll need to restart JIRA after you're done with the setup");
			}
			return false;
		}
		
		if (LOGIN_JSP.equals(path)) {
			if (LOG.isTraceEnabled()) {
				LOG.trace("request " + _req.getServletPath() + " is login.jsp");
			}
			return false;
		}

		final String referer = _req.getHeader(REFERER);
		if (referer != null && referer.endsWith(LOGIN_JSP)) {
			if (LOG.isTraceEnabled()) {
				LOG.trace("request " + _req.getServletPath() + " is referred by " + referer);
			}
			return false;
		}
		
		if (LOG.isTraceEnabled()) {
			LOG.trace("isAuthenticationRequired: yes");
		}
		
		return true;
	}


	protected void postAuthentication(final boolean _success) {
		if (!_success) {
			if (LOG.isErrorEnabled()) {
				LOG.error("NTLM went ok but could not login to JIRA.");
				LOG.error("This could be either a bad login, or you might have configured a custom Authenticator for JIRA.");
				LOG.error("Check the README.");
			}

			//NTLMLoginFilter.isNTLMEnabled = false;
		}
	}


	/**
	 * Attempt a login, making sure the user is created first if necessary _and_ requested via config.
	 */
	public boolean doLogin(final HttpServletRequest _req, final HttpServletResponse _res, final String _domain, final String _username) {
		if (!LDAPConfig.allowDomain(_domain)) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("no configuration for domain " + _domain);
			}
			return false;
		}

		// We make sure the user exists in the local Jira storage.
		User user = JIRAUserUtils.getUser(_domain, _username);
		if (user == null) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("User " + _username + " not found in JIRA");
			}
			return false;
		}

		// Update user data from the LDAP.
		if (!JIRAUserUtils.updateUserData(user, _domain, _username)) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("Failed to update user data from LDAP");
			}
			return false;
		}

		return super.doLogin(_req, _res, _username.toLowerCase(), BypassDefaultAuthenticator.NO_PASSWORD);
	}
}
