/*
 * Author: Daniel Pavel <daniel.pavel@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name of the Softwin nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package sk.lacike.ntlm.jcifs.jira;

import java.io.IOException;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import sk.lacike.ntlm.LDAPConfig;

import jcifs.smb.NtStatus;
import jcifs.smb.NtlmPasswordAuthentication;
import jcifs.smb.SmbException;

public class NTLMWebHelper {
	private static final Log LOG = LogFactory.getLog(NTLMWebHelper.class);

	// used in headers for NTLM
	public static final String NTLM = "NTLM";
	public static final String NTLM_SPACE = NTLM + " ";
	public static final String AUTHORIZATION = "Authorization";
	public static final String WWW_AUTHENTICATE = "WWW-Authenticate";

	// we cache some jCIFS data in the servlet session
	public static final String SESSION_CHALLENGE = "NtlmHttpChal";
	public static final String SESSION_AUTHENTICATION = "NtlmHttpAuth";

	/**
	 *
	 * @param req
	 * @param res
	 * @throws IOException
	 * @throws ServletException
	 */
	public static NTLMResult doNTLM(final HttpServletRequest req, final HttpServletResponse res) throws IOException, ServletException {
		final NTLMResult result = new NTLMResult();

		final String authz = req.getHeader(AUTHORIZATION);
		if (authz == null || !authz.startsWith(NTLM_SPACE)) {
			// we tell the browser we want NTLM
			final HttpSession ssn = req.getSession(false);
			NtlmPasswordAuthentication ntlm = null;
			if (ssn != null) {
				ntlm = (NtlmPasswordAuthentication) ssn.getAttribute(SESSION_AUTHENTICATION);
			}
			if (ntlm == null) {
				res.setHeader(WWW_AUTHENTICATE, NTLM);
				res.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
				res.flushBuffer();
			}

			// we haven't actually _connected_ yet, but we need to tell back that auth is not yet done
			// and nothing went wrong (yet)
			return result;
		}

		result.gotFirstResponse = true;
		final byte[] challenge = SMBHelper.getDCChallenge();
		NtlmPasswordAuthentication pa = null;
		try {
			pa = SMBHelper.authenticate(req, res, challenge);
		} catch (SmbException smbex) {
			// can't connect to SMB, so we won't be able to property authenticate
			LOG.error("authenticate: no connection?", smbex);
			return result;
		}

		result.connected = true;
		if (pa == null) { // mid-handshake
			return result;
		}

		final HttpSession ssn = req.getSession();
		ssn.removeAttribute(SESSION_CHALLENGE);
		LOG.debug("tentative NTLM: [" + pa.getDomain() + "\\" + pa.getUsername() + "]");

		result.handshakeFinished = true;
		result.setUsername(pa.getUsername());
		/* lacike@lacike.sk - strip domain if necessary BEGIN */
		if (LDAPConfig.isStripDomain()) {
			LOG.debug("Stripping domain: [" + pa.getDomain() + "]");
			// set domain to empty string
			result.setDomain("");
			
			// clear domain also from username if present - sign @
			int idx = pa.getUsername().indexOf("@"); 
			if (idx > 0) {
				LOG.debug("Stripping domain from username: [" + pa.getUsername() + "]");
				result.setUsername(pa.getUsername().substring(0, idx - 1));
				LOG.debug("Result is: [" + pa.getUsername() + "]");
			}
			// clear domain also from username if present - sign $
			idx = pa.getUsername().indexOf("$"); 
			if (idx > 0) {
				LOG.debug("Stripping domain from username: [" + pa.getUsername() + "]");
				result.setUsername(pa.getUsername().substring(0, idx - 1));
				LOG.debug("Result is: [" + pa.getUsername() + "]");
			}
		} else
			result.setDomain(pa.getDomain());
		if (LDAPConfig.isForceDomain()) {
			LOG.debug("Forcing domain: [" + LDAPConfig.getForceDomain() + "]");
			result.setDomain(LDAPConfig.getForceDomain());
		}
		/* lacike@lacike.sk - strip domain if necessary END */
		if (!result.areIDsOK()) {
			return result;
		}
		
		if (LDAPConfig.isForceDomain() || LDAPConfig.isStripDomain()) {
			LOG.debug("Overriding NtlmPasswordAuthentication...");			
			pa = new NtlmPasswordAuthentication(
					result.getDomain(),
					result.getUsername(),
					challenge,
					pa.getAnsiHash(challenge),
					pa.getUnicodeHash(challenge));
		}

		try {
			SMBHelper.doDCLogon(pa);
			result.loginConfirmed = true;
		} catch (SmbException smbex) {
			/* Server challenge no longer valid for
			 * externally supplied password hashes.
			 */
			if (smbex.getNtStatus() == NtStatus.NT_STATUS_ACCESS_VIOLATION) {
				ssn.removeAttribute(SESSION_AUTHENTICATION);
			}

			// We could force ntlm again, but it may be best to give the user a chance to login the standard (form) way.
			//res.setHeader(WWW_AUTHENTICATE, NTLM);
			//res.setStatus(HttpServletResponse.SC_UNAUTHORIZED);
			//res.flushBuffer();
			if (LOG.isErrorEnabled()) {
				LOG.error("logon failed", smbex);
			}
		}

		return result;
	}

	/**
	 * Detect if IE is trying to re-negociate NTLM during a form POST request.
	 * @param _req the HTTP servlet request
	 */
	public static boolean isIERenegociatingPOST(final HttpServletRequest _req, final HttpServletResponse _res) {
		//final String remoteUser = NTLMLoginFilter.getRemoteUser(_req, _res);
		//if (remoteUser == null) {
		//	return false;
		//}

		final boolean yes = _req.getHeader(AUTHORIZATION) != null
			&& "POST".equals(_req.getMethod())
			&& _req.getContentLength() == 0
			&& _req.getHeader("User-Agent") != null
			&& _req.getHeader("User-Agent").indexOf("MSIE") > 0;
		if (yes) {
			if (LOG.isDebugEnabled()) {
				LOG.debug("IE renegotiating POST");
			}
		}

		return yes;
	}

	public static String defaultDomain = null;

	/**
	 * Container.  Holds NTLM authentication status.
	 * <p>
	 * The correct order in which you need to query the result state is:
	 * <ol>
	 * <li>barelyStarted()
	 * <li>hasFailedConnection()
	 * <li>hasFinishedHandshake()
	 * <li>hasFailedLogin() 
	 * </ol>
	 * Other than handling the failure cases (steps 2 and 4), your servlet should do nothing
	 * until the last step has finished (just <code>return</code>).
	 */
	public static class NTLMResult {
		boolean gotFirstResponse = false;
		// ntlm connection went ok
		boolean connected = false;
		// ntlm handshake finished
		boolean handshakeFinished = false;
		// ntlm login went ok
		boolean loginConfirmed = false;

		private String domain = null;
		private String username = null;

		/**
		 * First NTLM run initiated.
		 */
		public boolean barelyStarted() {
			return !gotFirstResponse;
		}

		/**
		 * We failed in connecting to the Domain Controller.
		 * No NTLM authentication will be possible, consider possible fallbacks.
		 */
		public boolean hasFailedConnection() {
			return gotFirstResponse && !connected;
		}

		/**
		 * The first NTLM run has finished successfully, and the second NTLM run was initiated.
		 */
		public boolean hasFinishedHandshake() {
			return !hasFailedConnection() && handshakeFinished;
		}

		/**
		 * The login procedure has finished, but the identity could not be confirmed.
		 */
		public boolean hasFailedLogin() {
			return hasFinishedHandshake() && (!loginConfirmed || !areIDsOK());
		}

		/**
		 * Valid username and domain were present in the NTLM response from the web browser.
		 */
		public boolean areIDsOK() {
			return domain != null && username != null;
		}

		/**
		 */
		public String getDomain() {
			return domain;
		}

		/**
		 * 
		 */
		public String getUsername() {
			return username;
		}

		/**
		 * @param _domain
		 */
		public void setDomain(final String _domain) {
			domain = _domain;

			if (domain == null || domain.length() == 0) {
				domain = defaultDomain;
			}

			// This solves a _whole_ lot of problems.
			if (domain != null) {
				domain = domain.toUpperCase();
			}
		}

		/**
		 * 
		 * @param _username
		 */
		public void setUsername(final String _username) {
			username = (_username == null || _username.length() == 0) ? null : _username;
		}
	}
}
