/*
 * Author: Daniel Pavel <daniel.pavel@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   * Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   * Neither the name of the Softwin nor the names of its contributors may
 *     be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package sk.lacike.ntlm.jcifs.jira;

import java.io.IOException;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.Enumeration;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import jcifs.http.NtlmSsp;
import jcifs.smb.NtStatus;
import jcifs.smb.NtlmPasswordAuthentication;
import jcifs.smb.SmbException;
import jcifs.smb.SmbSession;
import jcifs.UniAddress;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


public class SMBHelper {
	private static final Log LOG = LogFactory.getLog(SMBHelper.class);

	public static final int NETBIOS_SSN = 139;
	public static final int MICROSOFT_DS = 445;

	public static final UniAddress LOCALHOST_DC;
	public static final int LOCALHOST_DC_PORT;

	public static final String JCIFS_DEFAULT_DOMAIN_KEY = "jcifs.smb.client.domain";
	public static final String JCIFS_DEFAULT_NO_DOMAIN = "?";

	/*
	 * We need to find a network interface that responds to SMB requests.
	 */
	static {
		UniAddress foundInterf = null;
		int foundPort = -1;

		try {
			final Enumeration networks = NetworkInterface.getNetworkInterfaces();

			while (networks.hasMoreElements() && foundInterf == null) {
				final NetworkInterface n = (NetworkInterface) networks.nextElement();

				for (final Enumeration ei = n.getInetAddresses(); ei.hasMoreElements() && foundInterf == null; ) {
					final InetAddress ia = (InetAddress) ei.nextElement();

					if (ia.isLinkLocalAddress()) {
						continue;
					}

					// Costin Miroslav: also skip vmware virtual interfaces
					if (ia.isLoopbackAddress()) {
						continue;
					}

					final UniAddress u = new UniAddress(ia);
					final int port = findDCPort(u);
					if (port > 0) {
						foundPort = port;
						foundInterf = u;
					}
				}
			}
		} catch (SocketException sex) {
			LOG.error("Cannot enumerate network interfaces", sex);
		}

		LOCALHOST_DC = foundInterf;
		LOCALHOST_DC_PORT = foundPort;
		if (LOCALHOST_DC == null) {
			if (LOG.isWarnEnabled()) {
				LOG.warn("Could not connect to localhost's Windows authentication source!");
			}
		} else {
			if (LOG.isInfoEnabled()) {
				LOG.info("Localhost Controller available: " + LOCALHOST_DC + ":" + LOCALHOST_DC_PORT);
			}
		}
	}


	private static UniAddress domainController = null;
	private static int domainControllerPort = 0;

	protected static int findDCPort(final UniAddress _dc) {
		final int[] DCPorts = new int[] { NETBIOS_SSN, MICROSOFT_DS };

		for (int i = 0; i < DCPorts.length; i++) {
			try {
				if (LOG.isDebugEnabled()) {
					LOG.debug("Probing " + _dc + ":" + DCPorts[i] + " for a SMB connection...");
				}
				SmbSession.getChallenge(_dc, DCPorts[i]);
				return DCPorts[i];
			} catch (Exception ex) {
				if (LOG.isWarnEnabled()) {
					Throwable reason = ex;
					while (reason.getCause() != null) {
						reason = reason.getCause();
					}

					LOG.warn("SMB connection failed on InetAddress " + _dc, reason);
				}
			}
		}

		return -1;
	}


	/**
	 * Try to get a challenge from the domain controller.
	 */
	public static byte[] getDCChallenge() throws SmbException {
		try {
			return SmbSession.getChallenge(getDC(), getDCPort());
		} catch (UnknownHostException uhex) {
			// this makes no sense...
			if (LOG.isErrorEnabled()) {
				LOG.error("eh? getDCChallenge", uhex);
			}
			return null;
		}
	}


	/**
	 * Try the actual SMB logon.
	 * @see jcifs.smb.NtStatus
	 */
	public static void doDCLogon(final NtlmPasswordAuthentication _ntlm) throws SmbException {
		/*
		if (getDC() == null)
			throw new SmbException(NtStatus.NT_STATUS_INVALID_HANDLE);
		*/

		SmbSession.logon(getDC(), getDCPort(), _ntlm);
	}


	public static NtlmPasswordAuthentication authenticate(final HttpServletRequest _req, final HttpServletResponse _res, final byte[] _challenge)
			throws IOException, ServletException {

		final byte[] challenge = _challenge == null ? getDCChallenge() : _challenge;
		return NtlmSsp.authenticate(_req, _res, challenge);
	}


	/**
	 * Gets the human-readable NT error message from a NT_STATUS_ return code.
	 * @param _smbError NT error code
	 */
	public static String getErrorMessage(final int _smbError) {
		for (int i = 0; i < NtStatus.NT_STATUS_CODES.length; i++) {
			if (_smbError == NtStatus.NT_STATUS_CODES[i]) {
				return NtStatus.NT_STATUS_MESSAGES[i];
			}
		}

		return null;
	}


	public static boolean setDC(final String _address) throws UnknownHostException {
		LOG.debug("Setting Domain Controller " + domainController + ":" + domainControllerPort);
		if (_address == null) {
			domainController = null;
			domainControllerPort = 0;
			return true;
		}

		final UniAddress host = UniAddress.getByName(_address);
		int port = findDCPort(host);
		if (port < 0) {
			LOG.debug("Wrong Domain Controller port " + domainControllerPort);
			return false;
		}

		domainController = host;
		domainControllerPort = port;
		LOG.info("Using Domain Controller " + domainController + ":" + domainControllerPort);
		return true;
	}


	public static UniAddress getDC() {
		return domainController == null ? LOCALHOST_DC : domainController;
	}


	public static int getDCPort() {
		return domainController == null ? LOCALHOST_DC_PORT : domainControllerPort;
	}
}
